//八皇后问题的验证程序（递归实现）
// 要点：
//     1 归纳谓词描述皇后在棋盘上正确摆放的条件
//     2 本例有2个函数：函数checkq 检查第n个皇后是否和前n-1个皇后处于同列或同对角线。
//        函数 eightQueensRec 递归求解八皇后问题，并输出所有找到的解。
//     3 main函数作为测试函数调用eightQueensRec函数
// 关键字：递归函数、归纳谓词、存在量词

#include <stdio.h>
#include <stdlib.h>
#include <limits.h>

#define QUEENS 8  //定义结果数组的大小，也就是皇后的数目     
typedef int SRowCol[QUEENS+1]; 
SRowCol a={0};        // 八皇后问题的皇后所在的行列位置，下标表示行，相应下标变量的值表示列,
                  // a[i]=j 表示第i行的皇后放在第j列。皇后从a[1]开始放置
int count = 0;    // 当前找到的八皇后问题的解的数目

///////////////////////////////////////全局逻辑定义和声明/////////////////////////////////////////////
/*@ 
    // 归纳谓词，描述a[first..n]中没有皇后出现在同一列上
    inductive NotSameColumn(SRowCol a,int first, int n) =   
        n == first 
        || 
        n >= first+1 && NotSameColumn(a,first, n-1) 
        && (\forall int i:[first+1..n-1].a[i]!=a[n])
        ;
    // 归纳谓词，描述a[first..n]中没有皇后出现在同一条对角线上
    inductive NotSameDiagonal(SRowCol a,int first, int n) = 
        n == first 
        || 
        n >= first+1 && NotSameDiagonal(a,first, n-1) 
        && (\forall int i:[first+1..n-1].a[n]!=a[i]-(i-n) && a[n]!=a[i]+(i-n))
        ;
*/

/////////////////////////////////////checkq函数的验证///////////////////////////////////////
// 前条件：
//   给出数组a的大小，皇后和皇后所在列的取值范围，以及前n-1位皇后已满足要求
// 后条件：
//   返回测试的结果：
//     =1，不存在其他皇后与a[n]处在同一列和同一对角线上
//     =0，有皇后与a[n]处于同一列或者同一对角线上。
/*@ 
    requires 
        \length(a) == QUEENS+1 && n >= 1 && n <= QUEENS 
        && (\forall int j:[1..n].0 < a[j] <= QUEENS) 
        && NotSameColumn(a,0,n-1) && NotSameDiagonal(a,0,n-1)
        ; 
    ensures 
        \result == 1 && \length(a) == QUEENS+1  //不存在皇后同行、同列和同对角线
        && NotSameColumn(a,0,n) && NotSameDiagonal(a,0,n) 
        ||
        \result == 0 && \length(a) == QUEENS+1  //存在皇后与a[n]同列
        && (\exists int j:[1..n-1].a[n]==a[j]) 
        ||
        \result == 0 && \length(a) == QUEENS+1  //存在皇后与a[n]同对角线
        && (\exists int j:[1..n-1].a[n]==a[j]-(j-n) || a[n] == a[j]+(j-n))
        ;
*/
// 检查当前皇后n放置的位置是否合理 
int checkq(const int n)
{
    int i;
    
    /*@ loop invariant
        \length(a) == QUEENS+1 && 1 <= i <= n && n <= QUEENS 
        &&  NotSameColumn(a,0,n-1) && NotSameDiagonal(a,0,n-1) 
        && (\forall int i:[1..n].0 < a[i] <= QUEENS)
        && (\forall int k:[1..i-1].a[k]!=a[n] && a[n]!=a[k]-(k-n) && a[n]!=a[k]+(k-n))
        ;
    */
    for(i=1;i<=n-1;i++){
        if((a[n]==a[i]) || (a[n]== a[i]-(i-n)) || (a[n]== a[i] + (i-n))) {  
            return 0;
        }
    }
    return 1;
}

/////////////////////////////////////eightQueensRec函数的验证///////////////////////////////////////
// 前条件：
//   给出测试第k位皇后时的条件
// 后条件：
//    1. 未找到解。1）第k位皇后的列号已定，测试第k+1位皇后；否则，2）测试下一个列的位置
//    2. 找到一个解，输出结果
// exits
//    解的个数已超出最大整型数的表示范围，中断退出，状态值为1


/*@ 
    requires  
        //加入第k个皇后前的k-1个皇后不同列且不同对角线
        1 <= k <= QUEENS+1 && (\forall int i:[1..k-1].0 < a[i] <= QUEENS)  
        && NotSameColumn(a,0,k-1) && NotSameDiagonal(a,0,k-1)  
        && INT_MAX >= count >= 0
        ;
    assigns 
        a[k..QUEENS], count;
    ensures   
        1 <= k <= QUEENS && INT_MAX >= count >= 0  // 未达到解
        && (\forall int i:[1..k].0 < a[i] <= QUEENS)  
        && (NotSameColumn(a,0,k) && NotSameDiagonal(a,0,k)    //前k个皇后不同列且不同对角线
            || 
            (\exists int i:[2..k].\exists int j:[1..i-1].a[i]==a[j]) //前k个皇后间存在同列
            ||
            (\exists int i:[2..k].\exists int j:[1..i-1].
                       (a[i]==a[j]-(j-i)) || a[i] == a[j]+(j-i)) //前k个皇后间存在同一对角线 
        ) 
        ||
        k == QUEENS+1 && (\forall int j:[1..QUEENS].0 < a[j] <= QUEENS)  // 达到解
        && NotSameColumn(a,0,k-1) && NotSameDiagonal(a,0,k-1) 
        && INT_MAX >= count >= 0
        ;
    exits 
        \exit_status == 1    // count > INT_MAX 
        ;
*/
void eightQueensRec(const int k)
{
    int i = 0;
    if(k>QUEENS) { //找到一个解，输出之
        if(count == INT_MAX){
            exit(1);
        }
        count++;
        printf("Solutions no.%d : ", count);
        /*@ loop invariant  
            1 <= i <= QUEENS + 1 && k == QUEENS+1 && INT_MAX >= count >= 0 
            && NotSameColumn(a,0,k-1) && NotSameDiagonal(a,0,k-1) 
            && (\forall int i:[1..QUEENS].0 < a[i] <= QUEENS) 
            ; 
        */
        for(i = 1;i<=QUEENS;i++){
            printf(" (%d,%d) ",i,a[i]);
        }
        printf("\n");    
    } else{
        /*@ loop invariant 
            1 <= i <= QUEENS + 1 && 1 <= k <= QUEENS && INT_MAX >= count >= 0 
            && NotSameColumn(a,0,k-1) && NotSameDiagonal(a,0,k-1) 
            && 
            (i == 1 && (\forall int j:[1..k-1].0 < a[j] <= QUEENS) 
             || 
              i >= 2 && (\forall int j:[1..k].0 < a[j] <= QUEENS) 
              &&
              ( //第k个皇后和前k-1个皇后不同列且不同对角线
                ((\forall int i:[1..k-1].a[i]!=a[k] 
                                      && a[k]!=a[i]-(i-k) 
                                      && a[k]!=a[i]+(i-k)) 
                  && a[k] == i-1) 
                ||
                //第k个皇后和前k-1个皇后间存在同列
                ((\exists int i:[1..k-1].a[k]==a[i]) 
                  && a[k] == i-1)
                ||
                //第k个皇后和前k-1个皇后间存在同对角线
                ( (\exists int i:[1..k-1].a[k]==a[i]-(i-k)||a[k] == a[i]+(i-k)) 
                  && a[k]== i-1)
              ) 
            )
            ; 
        */
        for(i=1;i<=QUEENS;i++){
            a[k] = i;
            if (checkq(k) == 1)
                eightQueensRec(k+1);
        
        }
    }
}

//
/*@
    requires \length(a) == QUEENS+1 && count == 0; 
    assigns count,a[1..8];
*/
int main(int argc, char **argv)
{
    eightQueensRec(1);
    return 0;
}